/*
* Copyright 2009, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef MODEL_IO_H__ 
#define MODEL_IO_H__ 

#include "general-io.h"
#include "kytea-model.h"
#include "kytea-config.h"
#include "dictionary.h"

#define MODEL_IO_VERSION "0.02"

using namespace std;

namespace kytea {

class ModelIO : public GeneralIO {

public:

    typedef char Format;
    const static Format FORMAT_BINARY = 'B';
    const static Format FORMAT_TEXT = 'T';
    const static Format FORMAT_UNKNOWN = 'U';

public:

    ModelIO(StringUtil* util) : GeneralIO(util) { }
    ModelIO(StringUtil* util, const char* file, bool out, bool bin) : GeneralIO(util,file,out,bin) { }
    ModelIO(StringUtil* util, iostream & str, bool out, bool bin) : GeneralIO(util,str,out,bin) { }

    virtual ~ModelIO() { }

    static ModelIO* createIO(const char* file, Format form, bool output, KyteaConfig & config);
    static ModelIO* createIO(iostream & str, Format form, bool output, KyteaConfig & config);

    virtual void writeConfig(const KyteaConfig & conf) = 0;
    virtual void writeModel(const KyteaModel * mod) = 0;
    virtual void writeDictionary(const Dictionary * dict) = 0;

    virtual void readConfig(KyteaConfig & conf) = 0;
    virtual KyteaModel * readModel() = 0;
    virtual Dictionary * readDictionary() = 0;

};

class TextModelIO : public ModelIO {

public:

    TextModelIO(StringUtil* util) : ModelIO(util) { }
    TextModelIO(StringUtil* util, const char* file, bool out) : ModelIO(util,file,out,false) { }
    TextModelIO(StringUtil* util, iostream & str, bool out) : ModelIO(util,str,out,false) { }

    void writeConfig(const KyteaConfig & conf);
    void writeModel(const KyteaModel * mod);
    void writeDictionary(const Dictionary * dict);

    // create an appropriate parser based on the type
    static CorpusIO* createIO(const char* file, Format form, bool output, StringUtil* util);
    static CorpusIO* createIO(iostream & str, Format form, bool output, StringUtil* util);

    void readConfig(KyteaConfig & conf);
    KyteaModel * readModel();
    Dictionary * readDictionary();

};

class BinaryModelIO : public ModelIO {

public:

    BinaryModelIO(StringUtil* util) : ModelIO(util) { }
    BinaryModelIO(StringUtil* util, const char* file, bool out) : ModelIO(util,file,out,true) { }
    BinaryModelIO(StringUtil* util, iostream & str, bool out) : ModelIO(util,str,out,true) { }

    void writeConfig(const KyteaConfig & conf);
    void writeModel(const KyteaModel * mod);
    void writeDictionary(const Dictionary * dict);

    void readConfig(KyteaConfig & conf);
    KyteaModel * readModel();
    Dictionary * readDictionary();

};

}

#endif
